from gae.layers import GraphConvolution, GraphConvolutionSparse, InnerProductDecoder
import tensorflow as tf
from gae.preprocessing import preprocess_graph
flags = tf.app.flags
FLAGS = flags.FLAGS


class Model(object):
    def __init__(self, **kwargs):
        allowed_kwargs = {'name', 'logging'}
        for kwarg in kwargs.keys():
            assert kwarg in allowed_kwargs, 'Invalid keyword argument: ' + kwarg

        for kwarg in kwargs.keys():
            assert kwarg in allowed_kwargs, 'Invalid keyword argument: ' + kwarg
        name = kwargs.get('name')
        if not name:
            name = self.__class__.__name__.lower()
        self.name = name

        logging = kwargs.get('logging', False)
        self.logging = logging

        self.vars = {}

    def _build(self):
        raise NotImplementedError

    def build(self):
        """ Wrapper for _build() """
        with tf.variable_scope(self.name):
            self._build()
        variables = tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES, scope=self.name)
        self.vars = {var.name: var for var in variables}

    def fit(self):
        pass

    def predict(self):
        pass


class GCNModelAE(Model):
    def __init__(self, placeholders, num_features, features_nonzero, **kwargs):
        super(GCNModelAE, self).__init__(**kwargs)

        self.inputs = placeholders['features']
        self.input_dim = num_features
        self.features_nonzero = features_nonzero
        self.adj = placeholders['adj']
        self.dropout = placeholders['dropout']
        self.build()

    def _build(self):

        self.hidden1 = GraphConvolutionSparse(input_dim=self.input_dim,
                                              output_dim=FLAGS.hidden1,
                                              adj=self.adj,
                                              features_nonzero=self.features_nonzero,
                                              act=tf.nn.relu,
                                              dropout=self.dropout,
                                              logging=self.logging)(self.inputs)

        self.embeddings = GraphConvolution(input_dim=FLAGS.hidden1,
                                           output_dim=FLAGS.hidden2,
                                           adj=self.adj,
                                           act=lambda x: x,
                                           dropout=self.dropout,
                                           logging=self.logging)(self.hidden1)

        self.z_mean = self.embeddings

        self.reconstructions = InnerProductDecoder(input_dim=FLAGS.hidden2,
                                      act=lambda x: x,
                                      logging=self.logging)(self.embeddings)


class GCNModelVAE_encoder(Model):
    def __init__(self, placeholders, num_features, num_nodes, features_nonzero, **kwargs):
        super(GCNModelVAE_encoder, self).__init__(**kwargs)

        self.inputs = placeholders['features']
        self.input_dim = num_features
        self.features_nonzero = features_nonzero
        self.n_samples = num_nodes
        self.adj = placeholders['adj']
        self.dropout = placeholders['dropout']
        self.c=tf.eye(num_nodes)
        self.build()

    def _build(self):

        self.hidden1 = GraphConvolutionSparse(input_dim=self.input_dim,
                                              output_dim=FLAGS.hidden1,
                                              adj=self.adj,
                                              features_nonzero=self.features_nonzero,
                                              act=tf.nn.relu,
                                              dropout=self.dropout,
                                              logging=self.logging)(self.inputs)

        self.z_mean = GraphConvolution(input_dim=FLAGS.hidden1,
                                       output_dim=FLAGS.hidden2,
                                       adj=self.adj,
                                       act=lambda x: x,
                                       dropout=self.dropout,
                                       logging=self.logging)(self.hidden1)

        self.z_log_std = GraphConvolution(input_dim=FLAGS.hidden1,
                                          output_dim=FLAGS.hidden2,
                                          adj=self.adj,
                                          act=lambda x: x,
                                          dropout=self.dropout,
                                          logging=self.logging)(self.hidden1)

        self.z = self.z_mean + tf.random_normal([self.n_samples, FLAGS.hidden2]) * tf.exp(self.z_log_std)



class GCNModelVAE_decoder(Model):
    def __init__(self, placeholders, num_features,  features_nonzero, **kwargs):
        super(GCNModelVAE_decoder, self).__init__(**kwargs)
        self.z = placeholders['z']
        self.n_samples = placeholders['num_nodes']
        self.adj = placeholders['adj']
        self.dropout = placeholders['dropout']
        self.c=tf.eye(placeholders['num_nodes'])
        self.F_mean_true = placeholders['F_mean_true']
        self.F_log_std_true = placeholders['F_log_std_true']
        self.build()

    def _build(self):


        #self.z= tf.random_normal([self.n_samples_out, FLAGS.hidden2])

        self.reconstructions = InnerProductDecoder(input_dim=FLAGS.hidden2,
                                      act=lambda x: x,
                                      logging=self.logging)(self.z)

        self.reconstructions_adj = tf.reshape(tf.cast(tf.greater_equal(tf.sigmoid(self.reconstructions), FLAGS.threshold), tf.float32), [self.n_samples,self.n_samples])

        adj_ = self.reconstructions_adj+ self.c
        rowsum=tf.reduce_sum(adj_,1)
        degree_mat_inv_sqrt = tf.matrix_diag(tf.pow(rowsum, -0.5))

        adj_normalized = tf.tensordot(tf.transpose(tf.tensordot(adj_,degree_mat_inv_sqrt,axes=1)),degree_mat_inv_sqrt,axes=1)
        adj_normalized=tf.sparse.from_dense(adj_normalized)

        self.hidden1_F = GraphConvolution(input_dim=FLAGS.hidden2,
                                          output_dim=FLAGS.hidden1,
                                          adj=adj_normalized,
                                          act=tf.nn.relu,
                                          dropout=self.dropout,
                                          logging=self.logging)(self.z)

        self.F_mean = GraphConvolution(input_dim=FLAGS.hidden1,
                                       output_dim=1,
                                       adj=adj_normalized,
                                       act=lambda x: x,
                                       dropout=self.dropout,
                                       logging=self.logging)(self.hidden1_F)

        self.F_log_std = GraphConvolution(input_dim=FLAGS.hidden1,
                                          output_dim=1,
                                          adj=adj_normalized,
                                          act=lambda x: x,
                                          dropout=self.dropout,
                                          logging=self.logging)(self.hidden1_F)

        self.F = self.F_mean + tf.random_normal([self.n_samples, 1]) * tf.exp(self.F_log_std)